from os import path
import sys

# Setup Django with our test demo project. We need to do this in global
# module code rather than setup_package(), because we want it to run
# before any module-wide imports in any of the test modules.
sys.path.insert(0, path.join(path.dirname(__file__), 'res', 'apps'))
from django.core.management import setup_environ
import settings
setup_environ(settings)



import django.views.generic.base as _generic_base
from coffin.template.response import TemplateResponse as JinjaTemplateResponse

class TemplateResponseMixin(_generic_base.TemplateResponseMixin):
    """
    A mixin that can be used to render a template using Jinja.
    """
    response_class = JinjaTemplateResponse

class TemplateView(TemplateResponseMixin, _generic_base.TemplateView):
    """
    A view that renders a template using Jinja.
    """


from django.contrib.syndication.views import *       # merge modules

import sys
from django.contrib.syndication.views import Feed as DjangoFeed
from coffin.template import loader as coffin_loader


class Feed(DjangoFeed):
    """A ``Feed`` implementation that renders it's title and
    description templates using Jinja2.

    Unfortunately, Django's base ``Feed`` class is not very extensible
    in this respect at all. For a real solution, we'd have to essentially
    have to duplicate the whole class. So for now, we use this terrible
    non-thread safe hack.

    Another, somewhat crazy option would be:
        * Render the templates ourselves through Jinja2 (possible
          introduce new attributes to avoid having to rewrite the
          existing ones).
        * Make the rendered result available to Django/the superclass by
          using a custom template loader using a prefix, say
          "feed:<myproject.app.views.MyFeed>". The loader would simply
          return the Jinja-rendered template (escaped), the Django template
          mechanism would find no nodes and just pass the output through.
    Possible even worse than this though.
    """

    def get_feed(self, *args, **kwargs):
        parent_module = sys.modules[DjangoFeed.__module__]
        old_loader = parent_module.loader
        parent_module.loader = coffin_loader
        try:
            return super(Feed, self).get_feed(*args, **kwargs)
        finally:
            parent_module.loader = old_loader


from coffin import template
from django.contrib.staticfiles.storage import staticfiles_storage
from coffin.templatetags.static import StaticExtension


register = template.Library()


class StaticExtension(StaticExtension):
    """Implements the {% static %} tag as provided by the ``staticfiles``
    contrib module.

    Rreturns the URL to a file using staticfiles' storage backend.

    Usage::

        {% static path [as varname] %}

    Examples::

        {% static "myapp/css/base.css" %}
        {% static variable_with_path %}
        {% static "myapp/css/base.css" as admin_base_css %}
        {% static variable_with_path as varname %}

    """

    @classmethod
    def get_statc_url(cls, path):
        return super(StaticExtension, cls).get_statc_url(
            staticfiles_storage.url(path))


register.tag(StaticExtension)


def static(path):
    return StaticExtension.get_static_url(path)



from django.template import Library as DjangoLibrary, InvalidTemplateLibrary
from jinja2.ext import Extension as Jinja2Extension
import types
from coffin.interop import (
    DJANGO, JINJA2,
    guess_filter_type, jinja2_filter_to_django, django_filter_to_jinja2)


__all__ = ['Library']


class Library(DjangoLibrary):
    """Version of the Django ``Library`` class that can handle both
    Django template engine tags and filters, as well as Jinja2
    extensions and filters.

    Tries to present a common registration interface to the extension
    author, but provides both template engines with only those
    components they can support.

    Since custom Django tags and Jinja2 extensions are two completely
    different beasts, they are handled completely separately. You can
    register custom Django tags as usual, for example:

        register.tag('current_time', do_current_time)

    Or register a Jinja2 extension like this:

        register.tag(CurrentTimeNode)

    Filters, on the other hand, work similarily in both engines, and
    for the most one can't tell whether a filter function was written
    for Django or Jinja2. A compatibility layer is used to make to
    make the filters you register usuable with both engines:

        register.filter('cut', cut)

    However, some of the more powerful filters just won't work in
    Django, for example if more than one argument is required, or if
    context- or environmentfilters are used. If ``cut`` in the above
    example where such an extended filter, it would only be registered
    with Jinja.

    See also the module documentation for ``coffin.interop`` for
    information on some of the limitations of this conversion.

    TODO: Jinja versions of the ``simple_tag`` and ``inclusion_tag``
    helpers would be nice, though since custom tags are not needed as
    often in Jinja, this is not urgent.
    """

    def __init__(self):
        super(Library, self).__init__()
        self.jinja2_filters = {}
        self.jinja2_extensions = []
        self.jinja2_environment_attrs = {}
        self.jinja2_globals = {}
        self.jinja2_tests = {}

    @classmethod
    def from_django(cls, django_library):
        """Create a Coffin library object from a Django library.

        Specifically, this ensures that filters already registered
        with the Django library are also made available to Jinja,
        where applicable.
        """
        from copy import copy
        result = cls()
        result.tags = copy(django_library.tags)
        for name, func in django_library.filters.iteritems():
            result._register_filter(name, func, type='django')
        return result

    def test(self, name=None, func=None):
        def inner(f):
            name = getattr(f, "_decorated_function", f).__name__
            self.jinja2_tests[name] = f
            return f
        if name == None and func == None:
            # @register.test()
            return inner
        elif func == None:
            if (callable(name)):
                # register.test()
                return inner(name)
            else:
                # @register.test('somename') or @register.test(name='somename')
                def dec(func):
                    return self.test(name, func)
                return dec
        elif name != None and func != None:
            # register.filter('somename', somefunc)
            self.jinja2_tests[name] = func
            return func
        else:
            raise InvalidTemplateLibrary("Unsupported arguments to "
                "Library.test: (%r, %r)", (name, func))

    def object(self, name=None, func=None):
        def inner(f):
            name = getattr(f, "_decorated_function", f).__name__
            self.jinja2_globals[name] = f
            return f
        if name == None and func == None:
            # @register.object()
            return inner
        elif func == None:
            if (callable(name)):
                # register.object()
                return inner(name)
            else:
                # @register.object('somename') or @register.object(name='somename')
                def dec(func):
                    return self.object(name, func)
                return dec
        elif name != None and func != None:
            # register.object('somename', somefunc)
            self.jinja2_globals[name] = func
            return func
        else:
            raise InvalidTemplateLibrary("Unsupported arguments to "
                "Library.object: (%r, %r)", (name, func))

    def tag(self, name_or_node=None, compile_function=None, environment={}):
        """Register a Django template tag (1) or Jinja 2 extension (2).

        For (1), supports the same invocation syntax as the original
        Django version, including use as a decorator.

        For (2), since Jinja 2 extensions are classes (which can't be
        decorated), and have the tag name effectively built in, only the
        following syntax is supported:

            register.tag(MyJinjaExtensionNode)

        If your extension needs to be configured by setting environment
        attributes, you can can pass key-value pairs via ``environment``.
        """
        if isinstance(name_or_node, type) and issubclass(name_or_node, Jinja2Extension):
            if compile_function:
                raise InvalidTemplateLibrary('"compile_function" argument not supported for Jinja2 extensions')
            self.jinja2_extensions.append(name_or_node)
            self.jinja2_environment_attrs.update(environment)
            return name_or_node
        else:
            if environment:
                raise InvalidTemplateLibrary('"environment" argument not supported for Django tags')
            return super(Library, self).tag(name_or_node, compile_function)

    def tag_function(self, func_or_node):
        if not isinstance(func_or_node, types.FunctionType) and \
                issubclass(func_or_node, Jinja2Extension):
            self.jinja2_extensions.append(func_or_node)
            return func_or_node
        else:
            return super(Library, self).tag_function(func_or_node)

    def filter(self, name=None, filter_func=None, type=None, jinja2_only=None):
        """Register a filter with both the Django and Jinja2 template
        engines, if possible - or only Jinja2, if ``jinja2_only`` is
        specified. ``jinja2_only`` does not affect conversion of the
        filter if neccessary.

        Implements a compatibility layer to handle the different
        auto-escaping approaches transparently. Extended Jinja2 filter
        features like environment- and contextfilters are however not
        supported in Django. Such filters will only be registered with
        Jinja.

        If you know which template language the filter was written for,
        you may want to specify type="django" or type="jinja2", to disable
        the interop layer which in some cases might not be able to operate
        entirely opaque. For example, Jinja 2 filters may not receive a
        "Undefined" value if the interop layer is applied.

        Supports the same invocation syntax as the original Django
        version, including use as a decorator.

        If the function is supposed to return the registered filter
        (by example of the superclass implementation), but has
        registered multiple filters, a tuple of all filters is
        returned.
        """
        def filter_function(f):
            return self._register_filter(
                getattr(f, "_decorated_function", f).__name__,
                f, type=type, jinja2_only=jinja2_only)
        if name == None and filter_func == None:
            # @register.filter()
            return filter_function
        elif filter_func == None:
            if (callable(name)):
                # @register.filter
                return filter_function(name)
            else:
                # @register.filter('somename') or @register.filter(name='somename')
                def dec(func):
                    return self.filter(name, func, type=type,
                                       jinja2_only=jinja2_only)
                return dec
        elif name != None and filter_func != None:
            # register.filter('somename', somefunc)
            return self._register_filter(name, filter_func, type=type,
                jinja2_only=jinja2_only)
        else:
            raise InvalidTemplateLibrary("Unsupported arguments to "
                "Library.filter: (%r, %r)", (name, filter_func))

    def jinja2_filter(self, *args, **kwargs):
        """Shortcut for filter(type='jinja2').
        """
        kw = {'type': JINJA2}
        kw.update(kwargs)
        return self.filter(*args, **kw)

    def _register_filter(self, name, func, type=None, jinja2_only=None):
        assert type in (None, JINJA2, DJANGO,)

        # The user might not specify the language the filter was written
        # for, but sometimes we can auto detect it.
        filter_type, can_be_ported = guess_filter_type(func)
        assert not (filter_type and type) or filter_type == type, \
               "guessed filter type (%s) not matching claimed type (%s)" % (
                   filter_type, type,
               )
        if not filter_type and type:
            filter_type = type

        if filter_type == JINJA2:
            self.jinja2_filters[name] = func
            if can_be_ported and not jinja2_only:
                self.filters[name] = jinja2_filter_to_django(func)
            return func
        elif filter_type == DJANGO:
            self.filters[name] = func
            if not can_be_ported and jinja2_only:
                raise ValueError('This filter cannot be ported to Jinja2.')
            if can_be_ported:
                self.jinja2_filters[name] = django_filter_to_jinja2(func)
            return func
        else:
            django_func = jinja2_filter_to_django(func)
            jinja2_func = django_filter_to_jinja2(func)
            if jinja2_only:
                self.jinja2_filters[name] = jinja2_func
                return jinja2_func
            else:
                # register the filter with both engines
                self.filters[name] = django_func
                self.jinja2_filters[name] = jinja2_func
                return (django_func, jinja2_func)



"""Compatibility functions between Jinja2 and Django.

General notes:

  - The Django ``stringfilter`` decorator is supported, but should not be
    used when writing filters specifically for Jinja: It will lose the
    attributes attached to the filter function by Jinja's
    ``environmentfilter`` and ``contextfilter`` decorators, when used
    in the wrong order.

    Maybe coffin should provide a custom version of stringfilter.

  - While transparently converting filters between Django and Jinja works
    for the most part, there is an issue with Django's
    ``mark_for_escaping``, as Jinja does not support a similar mechanism.
    Instead, for Jinja, we escape such strings immediately (whereas Django
    defers it to the template engine).
"""

import inspect
from django.utils.safestring import SafeUnicode, SafeData, EscapeData
from jinja2 import Markup, environmentfilter, Undefined


__all__ = (
    'DJANGO', 'JINJA2',
    'django_filter_to_jinja2',
    'jinja2_filter_to_django',
    'guess_filter_type',)


DJANGO = 'django'
JINJA2 = 'jinja2'


def django_filter_to_jinja2(filter_func):
    """
    Note: Due to the way this function is used by
    ``coffin.template.Library``, it needs to be able to handle native
    Jinja2 filters and pass them through unmodified. This necessity
    stems from the fact that it is not always possible to determine
    the type of a filter.

    TODO: Django's "func.is_safe" is not yet handled
    """
    def _convert_out(v):
        if isinstance(v, SafeData):
            return Markup(v)
        if isinstance(v, EscapeData):
            return Markup.escape(v)       # not 100% equivalent, see mod docs
        return v
    def _convert_in(v):
        if isinstance(v, Undefined):
            # Essentially the TEMPLATE_STRING_IF_INVALID default
            # setting. If a non-default is set, Django wouldn't apply
            # filters. This is something that we neither can nor want to
            # simulate in Jinja.
            return ''
        return v
    def conversion_wrapper(value, *args, **kwargs):
        result = filter_func(_convert_in(value), *args, **kwargs)
        return _convert_out(result)
    # Jinja2 supports a similar machanism to Django's
    # ``needs_autoescape`` filters: environment filters. We can
    # thus support Django filters that use it in Jinja2 with just
    # a little bit of argument rewriting.
    if hasattr(filter_func, 'needs_autoescape'):
        @environmentfilter
        def autoescape_wrapper(environment, *args, **kwargs):
            kwargs['autoescape'] = environment.autoescape
            return conversion_wrapper(*args, **kwargs)
        return autoescape_wrapper
    else:
        return conversion_wrapper


def jinja2_filter_to_django(filter_func):
    """
    Note: Due to the way this function is used by
    ``coffin.template.Library``, it needs to be able to handle native
    Django filters and pass them through unmodified. This necessity
    stems from the fact that it is not always possible to determine
    the type of a filter.
    """
    if guess_filter_type(filter_func)[0] == DJANGO:
        return filter_func
    def _convert(v):
        # TODO: for now, this is not even necessary: Markup strings have
        # a custom replace() method that is immume to Django's escape()
        # attempts.
        #if isinstance(v, Markup):
        #    return SafeUnicode(v)         # jinja is always unicode
        # ... Jinja does not have a EscapeData equivalent
        return v
    def wrapped(value, *args, **kwargs):
        result = filter_func(value, *args, **kwargs)
        return _convert(result)
    return wrapped


def guess_filter_type(filter_func):
    """Returns a 2-tuple of (type, can_be_ported).

    ``type`` is one of DJANGO, JINJA2, or ``False`` if the type can
    not be determined.

    ``can_be_ported`` is ``True`` if we believe the filter could be
    ported to the other engine, respectively, or ``False`` if we know
    it can't.

    TODO: May not yet use all possible clues, e.g. decorators like
    ``stringfilter``.
    TOOD: Needs tests.
    """
    if hasattr(filter_func, 'contextfilter') or \
       hasattr(filter_func, 'environmentfilter'):
            return JINJA2, False

    args = inspect.getargspec(filter_func)
    if len(args[0]) - (len(args[3]) if args[3] else 0) > 2:
        return JINJA2, False

    if hasattr(filter_func, 'needs_autoescape'):
        return DJANGO, True

    # Looks like your run of the mill Python function, which are
    # easily convertible in either direction.
    return False, True


from django.contrib.syndication.feeds import *       # merge modules

import sys
from django.contrib.syndication.feeds import Feed as DjangoDeprecatedFeed
from django.contrib.syndication.views import Feed as DjangoNewFeed
from coffin.template import loader as coffin_loader
from django import VERSION as DJANGO_VERSION


class Feed(DjangoDeprecatedFeed):
    """Django changed the syndication framework in 1.2. This class
    represents the old way, ported to Coffin. If you are using 1.2,
    you should use the ``Feed`` class in
    ``coffin.contrib.syndication.views``.

    See also there for some notes on what we are doing here.
    """

    def get_feed(self, *args, **kwargs):
        if DJANGO_VERSION < (1,2):
            parent_module = sys.modules[DjangoDeprecatedFeed.__module__]
        else:
            # In Django 1.2, our parent DjangoDeprecatedFeed class really
            # inherits from DjangoNewFeed, so we need to patch the loader
            # in a different module.
            parent_module = sys.modules[DjangoNewFeed.__module__]

        old_loader = parent_module.loader
        parent_module.loader = coffin_loader
        try:
            return super(Feed, self).get_feed(*args, **kwargs)
        finally:
            parent_module.loader = old_loader


from coffin.views.generic.base import TemplateResponseMixin
from coffin.views.generic.detail import SingleObjectTemplateResponseMixin
import django.views.generic.edit as _generic_edit


class FormView(TemplateResponseMixin, _generic_edit.BaseFormView):
    """
    Equivalent of django generic view FormView, but uses Jinja template renderer.
    """


class CreateView(SingleObjectTemplateResponseMixin, _generic_edit.BaseCreateView):
    """
    Equivalent of django generic view CreateView, but uses Jinja template renderer.
    """
    template_name_suffix = '_form'


class UpdateView(SingleObjectTemplateResponseMixin, _generic_edit.BaseUpdateView):
    """
    Equivalent of django generic view UpdateView, but uses Jinja template renderer.
    """
    template_name_suffix = '_form'


class DeleteView(SingleObjectTemplateResponseMixin, _generic_edit.BaseDeleteView):
    """
    Equivalent of django generic view DeleteView, but uses Jinja template renderer.
    """
    template_name_suffix = '_confirm_delete'



"""Test construction of the implicitly provided JinjaEnvironment,
in the common.py module.
"""

from coffin.common import get_env
from django.test.utils import override_settings


def test_i18n():
    with override_settings(USE_I18N=True):
        assert get_env().from_string('{{ _("test") }}').render() == 'test'


class TestLoaders:

    def test_django_loader_replace(self):
        from coffin.template.loaders import jinja_loader_from_django_loader
        from jinja2 import loaders

        # Test replacement of filesystem loader
        l = jinja_loader_from_django_loader('django.template.loaders.filesystem.Loader')
        assert isinstance(l, loaders.FileSystemLoader)

        # Since we don't do exact matches for the loader string, make sure we
        # are not replacing loaders that are outside the Django namespace.
        l = jinja_loader_from_django_loader('djangoaddon.template.loaders.filesystem.Loader')
        assert not isinstance(l, loaders.FileSystemLoader)

    def test_cached_loader(self):
        from jinja2 import loaders

        with override_settings(TEMPLATE_LOADERS=[
            ('django.template.loaders.cached.Loader', (
                'django.template.loaders.filesystem.Loader',
                'django.template.loaders.app_directories.Loader',
            )),]):
            env = get_env()
            assert len(env.loader.loaders) == 1
            cached_loader = get_env().loader.loaders[0]
            assert hasattr(cached_loader, 'template_cache')
            assert len(cached_loader.loader.loaders) == 2
            assert isinstance(cached_loader.loader.loaders[0], loaders.FileSystemLoader)

            # the cached loader can find a template too.
            assert env.loader.load(env, 'render-x.html').render({'x': 'foo'}) == 'foo'




import django.views.generic.detail as _generic_detail
from coffin.views.generic.base import TemplateResponseMixin as JinjaTemplateResponseMixin

class SingleObjectTemplateResponseMixin(JinjaTemplateResponseMixin, _generic_detail.SingleObjectTemplateResponseMixin):
    """
    Equivalent of django mixin SingleObjectTemplateResponseMixin, but uses Jinja template renderer.
    """

class DetailView(SingleObjectTemplateResponseMixin, _generic_detail.BaseDetailView):
    """
    Equivalent of django generic view DetailView, but uses Jinja template renderer.
    """



"""Jinja2's i18n functionality is not exactly the same as Django's.
In particular, the tags names and their syntax are different:

  1. The Django ``trans`` tag is replaced by a _() global.
  2. The Django ``blocktrans`` tag is called ``trans``.

(1) isn't an issue, since the whole ``makemessages`` process is based on
converting the template tags to ``_()`` calls. However, (2) means that
those Jinja2 ``trans`` tags will not be picked up my Django's
``makemessage`` command.

There aren't any nice solutions here. While Jinja2's i18n extension does
come with extraction capabilities built in, the code behind ``makemessages``
unfortunately isn't extensible, so we can:

  * Duplicate the command + code behind it.
  * Offer a separate command for Jinja2 extraction.
  * Try to get Django to offer hooks into makemessages().
  * Monkey-patch.

We are currently doing that last thing. It turns out there we are lucky
for once: It's simply a matter of extending two regular expressions.
Credit for the approach goes to:
http://stackoverflow.com/questions/2090717/getting-translation-strings-for-jinja2-templates-integrated-with-django-1-x
"""

import re
from django.core.management.commands import makemessages
from django.utils.translation import trans_real
from django.template import BLOCK_TAG_START, BLOCK_TAG_END

strip_whitespace_right = re.compile(r"(%s-?\s*(trans|pluralize).*?-%s)\s+" % (BLOCK_TAG_START, BLOCK_TAG_END), re.U)
strip_whitespace_left = re.compile(r"\s+(%s-\s*(endtrans|pluralize).*?-?%s)" % (BLOCK_TAG_START, BLOCK_TAG_END), re.U)

def strip_whitespaces(src):
    src = strip_whitespace_left.sub(r'\1', src)
    src = strip_whitespace_right.sub(r'\1', src)
    return src

class Command(makemessages.Command):

    def handle(self, *args, **options):
        old_endblock_re = trans_real.endblock_re
        old_block_re = trans_real.block_re
        old_templatize = trans_real.templatize
        # Extend the regular expressions that are used to detect
        # translation blocks with an "OR jinja-syntax" clause.
        trans_real.endblock_re = re.compile(
            trans_real.endblock_re.pattern + '|' + r"""^-?\s*endtrans\s*-?$""")
        trans_real.block_re = re.compile(
            trans_real.block_re.pattern + '|' + r"""^-?\s*trans(?:\s+(?!'|")(?=.*?=.*?)|-?$)""")
        trans_real.plural_re = re.compile(
            trans_real.plural_re.pattern + '|' + r"""^-?\s*pluralize(?:\s+.+|-?$)""")

        def my_templatize(src, origin=None):
            new_src = strip_whitespaces(src)
            return old_templatize(new_src, origin)

        trans_real.templatize = my_templatize

        try:
            super(Command, self).handle(*args, **options)
        finally:
            trans_real.endblock_re = old_endblock_re
            trans_real.block_re = old_block_re
            trans_real.templatize = old_templatize



try:
    from urllib.parse import urljoin
except ImportError:     # Python 2
    from urlparse import urljoin

from coffin.template import Library
from jinja2.ext import Extension
from jinja2 import nodes
from django.utils.encoding import iri_to_uri


register = Library()


class PrefixExtension(Extension):

    def parse(self, parser):
        stream = parser.stream
        lineno = stream.next().lineno

        call_node = self.call_method('render')

        if stream.next_if('name:as'):
            var = nodes.Name(stream.expect('name').value, 'store')
            return nodes.Assign(var, call_node).set_lineno(lineno)
        else:
            return nodes.Output([call_node]).set_lineno(lineno)

    def render(self, name):
        raise NotImplementedError()

    @classmethod
    def get_uri_setting(cls, name):
        try:
            from django.conf import settings
        except ImportError:
            prefix = ''
        else:
            prefix = iri_to_uri(getattr(settings, name, ''))
        return prefix


class GetStaticPrefixExtension(PrefixExtension):
    """
    Populates a template variable with the static prefix,
    ``settings.STATIC_URL``.

    Usage::

        {% get_static_prefix [as varname] %}

    Examples::

        {% get_static_prefix %}
        {% get_static_prefix as static_prefix %}

    """

    tags = set(['get_static_prefix'])

    def render(self):
        return self.get_uri_setting('STATIC_URL')


class GetMediaPrefixExtension(PrefixExtension):
    """
    Populates a template variable with the media prefix,
    ``settings.MEDIA_URL``.

    Usage::

        {% get_media_prefix [as varname] %}

    Examples::

        {% get_media_prefix %}
        {% get_media_prefix as media_prefix %}

    """

    tags = set(['get_media_prefix'])

    def render(self):
        return self.get_uri_setting('STATIC_URL')


class StaticExtension(PrefixExtension):
    """
    Joins the given path with the STATIC_URL setting.

    Usage::

        {% static path [as varname] %}

    Examples::

        {% static "myapp/css/base.css" %}
        {% static variable_with_path %}
        {% static "myapp/css/base.css" as admin_base_css %}
        {% static variable_with_path as varname %}

    """

    tags = set(['static'])

    def parse(self, parser):
        stream = parser.stream
        lineno = stream.next().lineno

        path = parser.parse_expression()
        call_node = self.call_method('get_statc_url', args=[path])

        if stream.next_if('name:as'):
            var = nodes.Name(stream.expect('name').value, 'store')
            return nodes.Assign(var, call_node).set_lineno(lineno)
        else:
            return nodes.Output([call_node]).set_lineno(lineno)

    @classmethod
    def get_statc_url(cls, path):
        return urljoin(PrefixExtension.get_uri_setting("STATIC_URL"), path)


register.tag(GetStaticPrefixExtension)
register.tag(GetMediaPrefixExtension)
register.tag(StaticExtension)


def static(path):
    return StaticExtension.get_static_url(path)



import re
from jinja2 import loaders

match_loader = re.compile(r'^(django|coffin)\.')


def jinja_loader_from_django_loader(django_loader, args=None):
    """Attempts to make a conversion from the given Django loader to an
    similarly-behaving Jinja loader.

    :param django_loader: Django loader module string.
    :return: The similarly-behaving Jinja loader, or None if a similar loader
        could not be found.
    """
    if not match_loader.match(django_loader):
        return None
    for substr, func in _JINJA_LOADER_BY_DJANGO_SUBSTR.iteritems():
        if substr in django_loader:
            return func(*(args or []))
    return None


def _make_jinja_app_loader():
    """Makes an 'app loader' for Jinja which acts like
    :mod:`django.template.loaders.app_directories`.
    """
    from django.template.loaders.app_directories import app_template_dirs
    return loaders.FileSystemLoader(app_template_dirs)


def _make_jinja_filesystem_loader():
    """Makes a 'filesystem loader' for Jinja which acts like
    :mod:`django.template.loaders.filesystem`.
    """
    from django.conf import settings
    return loaders.FileSystemLoader(settings.TEMPLATE_DIRS)


def _make_jinja_cached_loader(*loaders):
    """Makes a loader for Jinja which acts like
    :mod:`django.template.loaders.cached`.
    """
    return JinjaCachedLoader(
        [jinja_loader_from_django_loader(l) for l in loaders])


# Determine loaders from Django's conf.
_JINJA_LOADER_BY_DJANGO_SUBSTR = { # {substr: callable, ...}
    'app_directories': _make_jinja_app_loader,
    'filesystem': _make_jinja_filesystem_loader,
    'cached': _make_jinja_cached_loader,
    'AppLoader': _make_jinja_app_loader,
    'FileSystemLoader': _make_jinja_filesystem_loader,
}


class JinjaCachedLoader(loaders.BaseLoader):
    """A "sort of" port of of Django's "cached" template loader
    to Jinja 2. It exists primarily to support Django's full
    TEMPLATE_LOADERS syntax.

    However, note that it does not behave exactly like Django's cached
    loader: Rather than caching the compiled template, it only caches
    the template source, and recompiles the template every time. This is
    due to the way the Jinja2/Coffin loader setup works: The ChoiceLoader,
    which Coffin uses at the root to select from any of the configured
    loaders, calls the ``get_source`` method of each loader directly,
    bypassing ``load``. Our loader can therefore only hook into the process
    BEFORE template compilation.
    Caching the compiled templates by implementing ``load`` would only
    work if this loader instance were the root loader. See also the comments
    in Jinja2's BaseLoader class.

    Note that Jinja2 has an environment-wide bytecode cache (i.e. it caches
    compiled templates), that can function alongside with this class.

    Note further that Jinja2 has an environment-wide template cache (via the
    ``auto_reload`` environment option), which duplicate the functionality
    of this class entirely, and should be preferred when possible.
    """

    def __init__(self, subloaders):
        self.loader = loaders.ChoiceLoader(subloaders)
        self.template_cache = {}

    def get_source(self, environment, template):
        key = (environment, template)
        if key not in self.template_cache:
            result = self.loader.get_source(environment, template)
            self.template_cache[key] = result
        return self.template_cache[key]



from datetime import datetime, date
from nose.tools import assert_raises


def r(s, context={}):
    from coffin.common import env
    return env.from_string(s).render(context)


def test_django_builtins_available():
    """Many filters have not been re-implemented specifically for
    Coffin, but instead the Django version is used through an
    interop-layer.

    Make sure that those are properly made available in Jinja2.
    """
    from coffin.template import defaultfilters
    assert not hasattr(defaultfilters, 'get_digit')  # has no port
    assert r('{{ "23475"|get_digit("2") }}') == '7'
    assert r('{{ unknown|get_digit("2") }}') == ''


def test_jinja2_builtins():
    """Ensure that the Jinja2 builtins are available, and take
    precedence over the Django builtins (which we automatically convert
    and install).
    """
    # Django's default filter only accepts one argument.
    assert r('{{ unknown|default("2", True) }}') == '2'


def test_url():
    # project name is optional
    assert r('{{ "urls_app.views.index"|url() }}') == '/url_test/'
    assert r('{{ "apps.urls_app.views.index"|url() }}') == '/url_test/'


def test_default():
    """We make the Jinja2 default filter behave like Django's without
    arguments, but still support Jinja2 extended syntax.
    """
    assert r('{{ foo|default("default") }}') == 'default'
    assert r('{{ foo|default("default") }}', {'foo': False}) == 'default'
    assert r('{{ foo|default("default", False) }}', {'foo': False}) == 'False'


def test_pluralize():
    assert r('vote{{ 0|pluralize }}') == 'votes'
    assert r('vote{{ 1|pluralize }}') == 'vote'
    assert r('class{{ 2|pluralize("es") }}') == 'classes'
    assert r('cand{{ 0|pluralize("y", "ies") }}') == 'candies'
    assert r('cand{{ 1|pluralize("y", "ies") }}') == 'candy'
    assert r('cand{{ 2|pluralize("y", "ies") }}') == 'candies'
    assert r('vote{{ [1,2,3]|pluralize }}') == 'votes'
    assert r('anonyme{{ 0|pluralize("r", "") }}') == 'anonyme'
    assert r('anonyme{{ 1|pluralize("r", "") }}') == 'anonymer'
    assert r('vote{{ 1|pluralize }}') == 'vote'
    assert_raises(TypeError, r, 'vote{{ x|pluralize }}', {'x': object()})
    assert_raises(ValueError, r, 'vote{{ x|pluralize }}', {'x': 'foo'})


def test_floatformat():
    assert r('{{ 1.3434|floatformat }}') == '1.3'
    assert r('{{ 1.3511|floatformat }}') == '1.4'
    assert r('{{ 1.3|floatformat(2) }}') == '1.30'
    assert r('{{ 1.30|floatformat(-3) }}') == '1.300'
    assert r('{{ 1.000|floatformat(3) }}') == '1.000'
    assert r('{{ 1.000|floatformat(-3) }}') == '1'
    assert_raises(ValueError, r, '{{ "foo"|floatformat(3) }}')
    assert_raises(ValueError, r, '{{ 4.33|floatformat("foo") }}')


def test_date_stuff():
    from coffin.common import env
    assert r('a{{ d|date("Y") }}b', {'d': date(2007, 01, 01)}) == 'a2007b'
    assert r('a{{ d|time("H") }}b', {'d': datetime(2007, 01, 01, 12, 01, 01)}) == 'a12b'
    # TODO: timesince, timeuntil

    # Make sure the date filters can handle unset values gracefully.
    # While generally we'd like to be explicit instead of hiding errors,
    # this is a particular case where it makes sense.
    for f in ('date', 'time', 'timesince', 'timeuntil'):
        assert r('a{{ d|%s }}b' % f) == 'ab'
        assert r('a{{ d|%s }}b' % f, {'d': None}) == 'ab'


from coffin.template import loader
from django.views.generic import date_based as _date_based
import functools


archive_index = functools.partial(_date_based.archive_index, template_loader=loader)
archive_year = functools.partial(_date_based.archive_year, template_loader=loader)
archive_month = functools.partial(_date_based.archive_month, template_loader=loader)
archive_week = functools.partial(_date_based.archive_week, template_loader=loader)
archive_daye = functools.partial(_date_based.archive_day, template_loader=loader)
archive_today = functools.partial(_date_based.archive_today, template_loader=loader)

object_detail = functools.partial(_date_based.object_detail, template_loader=loader)



from django.contrib.flatpages.models import FlatPage
from django.contrib.flatpages.views import DEFAULT_TEMPLATE
from django.shortcuts import get_object_or_404
from django.http import HttpResponse, HttpResponseRedirect
from django.conf import settings
from django.core.xheaders import populate_xheaders
from django.utils.safestring import mark_safe
from django.views.decorators.csrf import csrf_protect

from coffin.template import RequestContext, loader

# This view is called from FlatpageFallbackMiddleware.process_response
# when a 404 is raised, which often means CsrfViewMiddleware.process_view
# has not been called even if CsrfViewMiddleware is installed. So we need
# to use @csrf_protect, in case the template needs {% csrf_token %}.
@csrf_protect
def flatpage(request, url):
    """
    Flat page view.

    Models: `flatpages.flatpages`
    Templates: Uses the template defined by the ``template_name`` field,
        or `flatpages/default.html` if template_name is not defined.
    Context:
        flatpage
            `flatpages.flatpages` object
    """
    if not url.endswith('/') and settings.APPEND_SLASH:
        return HttpResponseRedirect("%s/" % request.path)
    if not url.startswith('/'):
        url = "/" + url
    f = get_object_or_404(FlatPage, url__exact=url, sites__id__exact=settings.SITE_ID)
    # If registration is required for accessing this page, and the user isn't
    # logged in, redirect to the login page.
    if f.registration_required and not request.user.is_authenticated():
        from django.contrib.auth.views import redirect_to_login
        return redirect_to_login(request.path)
    if f.template_name:
        t = loader.select_template((f.template_name, DEFAULT_TEMPLATE))
    else:
        t = loader.get_template(DEFAULT_TEMPLATE)

    # To avoid having to always use the "|safe" filter in flatpage templates,
    # mark the title and content as already safe (since they are raw HTML
    # content in the first place).
    f.title = mark_safe(f.title)
    f.content = mark_safe(f.content)

    c = RequestContext(request, {
        'flatpage': f,
    })
    response = HttpResponse(t.render(c))
    populate_xheaders(request, response, FlatPage, f.id)
    return response
